package geniusweb.boa.biddingstrategy;

import java.util.HashMap;

import geniusweb.actions.Action;
import geniusweb.bidspace.BidsWithUtility;
import geniusweb.boa.NegoState;
import geniusweb.profile.Profile;
import geniusweb.profile.utilityspace.LinearAdditive;

/**
 * * This is an abstract class used to implement a TimeDependentAgent Strategy
 * adapted from S. Shaheen Fatima Michael Wooldridge Nicholas R. Jennings
 * Optimal Negotiation Strategies for Agents with Incomplete Information
 * http://eprints.ecs.soton.ac.uk/6151/1/atal01.pdf
 * <p>
 * The default strategy was extended to enable the usage of opponent models.
 * <p>
 * Note that this agent is not fully equivalent to the theoretical model,
 * loading the domain may take some time, which may lead to the agent skipping
 * the first bid. A better implementation is GeniusTimeDependent_Offering.
 * <p>
 * The time-dependent utility target function is u(t) = k + (1.0 - k) *
 * Math.pow((t - startTime) / (1.0 - startTime), 1.0 / e)
 * <p>
 * 
 * Parameters:
 * <ul>
 * <li>e: the Concession factor e in the time-dependent utility target function.
 * Must be in [0,100]. Default is 1.
 * <li>k: the initial utility k in the time-dependent utility target function.
 * Must be in [0,1]. For default k = 0 the agent starts with a bid of maximum
 * utility
 * </ul>
 * <p>
 * Translation from Genius
 * negotiator.boaframework.offeringstrategy.other.GeniusTimeDependent_Offering
 * 
 * 
 */
public class TimeDepemdentBiddingStrategy implements BiddingStrategy {
	private boolean initialized = false;
	private Float e = 1f;
	private Float k = 0f;
	private BidsWithUtility bidSpace;
	private Float min;
	private float max;

	@Override
	public Action getAction(NegoState state) {
		if (!initialized) {
			init(state);
			initialized = true;
		}

		double time = state.getProgress();
		double utilityGoal;
		utilityGoal = p(time);

		// if there is no opponent model available
		if (opponentModel instanceof NoModel) {
			nextBid = negotiationSession.getOutcomeSpace()
					.getBidNearUtility(utilityGoal);
		} else {
			nextBid = omStrategy.getBid(outcomespace, utilityGoal);
		}
		return nextBid;

	}

	private void init(NegoState state) {
		HashMap<String, Object> params = state.getSettings().getParemeters();
		Profile profile = state.getProfile();
		if (!(profile instanceof LinearAdditive))
			throw new IllegalArgumentException(
					"Requires a LinearAdditive space but got "
							+ profile.getClass());

		this.bidSpace = new BidsWithUtility((LinearAdditive) profile);
		if (params.containsKey("e")) {
			this.e = (Float) params.get("e");
			if (this.e < 0 || this.k > 1)
				throw new IllegalArgumentException(
						"e must be in [0,100] but got " + this.e);
		}
		if (params.containsKey("k")) {
			this.k = (Float) params.get("k");
			if (this.k < 0 || this.k > 1)
				throw new IllegalArgumentException(
						"K must be in [0,1] but got " + this.k);

		}
		this.min = bidSpace.getRange().getMin().floatValue();
		this.max = bidSpace.getRange().getMax().floatValue();
		if (params.containsKey("min")) {
			this.min = (Float) params.get("min");
			// FIXME this value might be illegal
		}
		if (params.containsKey("max")) {
			this.max = (Float) params.get("max");
			// FIXME this value might be illegal
		}
	}

	/**
	 * From the paper:
	 * 
	 * A wide range of time dependent functions can be defined by varying the
	 * way in which f(t) is computed. However, functions must ensure that 0 <=
	 * f(t) <= 1, f(0) = k, and f(1) = 1.
	 * 
	 * That is, the offer will always be between the value range, at the
	 * beginning it will give the initial constant and when the deadline is
	 * reached, it will offer the reservation value.
	 */
	public double f(double t) {
		double ft = k + (1.0 - k)
				* Math.pow((t - startTime) / (1.0 - startTime), 1.0 / e);
		return ft;
	}

}
